import { ITUIStore, IOptions, Task } from '../interface/ITUIStore';
import { StoreName, NAME } from '../const';
import CallStore from './callStore';

export default class TUIStore implements ITUIStore {
  static instance: TUIStore;
  public task: Task;
  private storeMap: Partial<Record<StoreName, any>>;
  private timerId: number = -1;
  constructor() {
    this.storeMap = {
      [StoreName.CALL]: new CallStore(),
    };
    // todo 此处后续优化结构后调整
    this.task = {} as Task; // 保存监听回调列表
  }

  /**
   * 获取 TUIStore 实例
  */
  static getInstance() {
    if (!TUIStore.instance) {
      TUIStore.instance = new TUIStore();
    }
    return TUIStore.instance;
  }

  /**
   * UI 组件注册监听回调
   * @param {StoreName} storeName store 名称
   * @param {IOptions} options 监听信息
  */
  watch(storeName: StoreName, options: IOptions) {
    if (!this.task[storeName]) {
      this.task[storeName] = {};
    }
    const watcher = this.task[storeName];
    Object.keys(options).forEach((key) => {
      const callback = options[key];
      if (!watcher[key]) {
        watcher[key] = new Map();
      }
      watcher[key].set(callback, 1);
      // 注册监听后立即通知一次默认数据，避免UI层初始化调用 getData
      this.notify(storeName, key);
    });
  }

  /**
   * UI 取消组件监听回调
   * @param {StoreName} storeName store 名称
   * @param {IOptions} options 监听信息,包含需要取消的回掉等
   */
  unwatch(storeName: StoreName, options: IOptions) {
    // todo 该接口暂未支持，unwatch掉同一类，如仅传入store注销掉该store下的所有callback，同样options仅传入key注销掉该key下的所有callback
    // options的callback function为必填参数，后续修改
    if (!this.task[storeName]) {
      return;
    };
    // if (isString(options)) {
    //   // 移除所有的监听
    //   if (options === '*') {
    //     const watcher = this.task[storeName];
    //     Object.keys(watcher).forEach(key => {
    //       watcher[key].clear();
    //     });
    //   } else {
    //     console.warn(`${NAME.PREFIX}unwatch warning: options is ${options}`);
    //   }
    //   return;
    // }
    const watcher = this.task[storeName];
    Object.keys(options).forEach((key: string) => {
      watcher[key].delete(options[key]);
    });
  }

  /**
   * 通用 store 数据更新，messageList 的变更需要单独处理
   * @param {StoreName} storeName store 名称
   * @param {string} key 变更的 key
   * @param {unknown} data 变更的数据
  */
  update(storeName: StoreName, key: string, data: unknown) {
    this.storeMap[storeName]?.update(key, data);
    this.notify(storeName, key);
  }

  /**
   * 获取 Store 数据
   * @param {StoreName} storeName store 名称
   * @param {string} key 待获取的 key
  */
  getData(storeName: StoreName, key: string) {
    return this.storeMap[storeName]?.getData(key);
  }

  /**
   * UI 组件注册监听回调
   * @param {StoreName} storeName store 名称
   * @param {string} key 变更的 key
  */
  private notify(storeName: StoreName, key: string) {
    if (!this.task[storeName]) {
      return;
    }
    const watcher = this.task[storeName];
    if (watcher[key]) {
      const callbackMap = watcher[key];
      const data = this.getData(storeName, key);
      for (const [callback] of callbackMap.entries()) {
        callback.call(this, data);
      }
    }
  }

  public reset(storeName: StoreName, keyList: Array<string> = [], isNotificationNeeded = false) {
    if (storeName in this.storeMap) {
      const store = this.storeMap[storeName];
      // reset all
      if (keyList.length === 0) {
        keyList = Object.keys(store?.store);
      }
      store.reset(keyList);
      if (isNotificationNeeded) {
        keyList.forEach((key) => {
          this.notify(storeName, key);
        });
      }
    }
  }
  // 批量修改多个 key-value
  public updateStore(params: any, name?: StoreName): void {
    let storeName = name ? name : StoreName.CALL;
    Object.keys(params).forEach(key => {
      this.update(storeName, key, params[key]);
    });
  }
}
